"""Utilities to validate a regular grid of tiles."""

from dataclasses import dataclass

import numpy as np

from fractal_converters_tools.tile import Tile


def _first_if_allclose(values: list[float]) -> tuple[bool, float]:
    """Return the first value if all values are close."""
    if len(values) == 0:
        raise ValueError("Empty list of values")

    if np.allclose(values, values[0]):
        return True, values[0]
    return False, 0.0


def _find_grid_size(tiles: list[Tile], offset_x, offset_y) -> tuple[int, int]:
    """Find the grid size of a list of tiles."""
    x = [tile.top_l.x for tile in tiles]
    y = [tile.top_l.y for tile in tiles]

    num_x = int(max(x) // offset_x) + 1
    num_y = int(max(y) // offset_y) + 1
    return num_x, num_y

@dataclass
class GridSetup:
    """Grid setup for a list of tiles.

    Attributes:
        size_x (float): Size of each tile in the x direction.
        size_y (float): Size of each tile in the y direction.
        offset_x (float): Offset of each tile in the x direction.
        offset_y (float): Offset of each tile in the y direction.
        num_x (int): Number of tiles in the x direction.
        num_y (int): Number of tiles in the y direction.

    """

    length_x: float = 0.0
    length_y: float = 0.0
    offset_x: float = 0.0
    offset_y: float = 0.0
    num_x: int = 0
    num_y: int = 0

def check_if_regular_grid(tiles: list[Tile]) -> tuple[str | None, GridSetup]:
    """Find the grid size of a list of tiles."""
    if len(tiles) == 0:
        return "Empty list of tiles", GridSetup()

    if len(tiles) == 1:
        return "Only one tile", GridSetup()

    # ------------------------------------------
    # Test 1: Check if all lengths are the same
    # ------------------------------------------
    tiles_length_x = [bbox.bot_r.x - bbox.top_l.x for bbox in tiles]
    check, length_x = _first_if_allclose(tiles_length_x)
    if not check:
        all_lengths = np.unique(tiles_length_x)
        return f"Not all lengths are the same: {all_lengths}", GridSetup()

    tiles_length_y = [bbox.bot_r.y - bbox.top_l.y for bbox in tiles]
    check, length_y = _first_if_allclose(tiles_length_y)
    if not check:
        all_lengths = np.unique(tiles_length_y)
        return f"Not all lengths are the same: {all_lengths}", GridSetup()
    # ------------------------------------------
    # Test 2: Check if all offsets are the same
    # ------------------------------------------
    # Find the tiles offsets
    pos_top_l_x = [tile.top_l.x for tile in tiles]
    pos_top_l_x = np.sort(pos_top_l_x)
    offsets_x = np.diff(pos_top_l_x)
    offsets_x = offsets_x[offsets_x > 1e-6]
    check, offset_x = _first_if_allclose(offsets_x)
    if not check:
        # all x_offset are not the same
        unique_offsets = np.unique(offsets_x)
        return f"Not all x offsets are the same: {unique_offsets}", GridSetup()

    pos_top_l_y = [tile.top_l.y for tile in tiles]
    pos_top_l_y = np.sort(pos_top_l_y)
    offsets_y = np.diff(pos_top_l_y)
    offsets_y = offsets_y[offsets_y > 1e-6]
    check, offset_y = _first_if_allclose(offsets_y)
    if not check:
        # all y_offset are not the same
        unique_offsets = np.unique(offsets_y)
        return f"Not all y offsets are the same: {unique_offsets}", GridSetup()
    # ------------------------------------------
    # Test 3: Check the edge case where the grid is slanted
    # ------------------------------------------
    for tile in tiles[1:]:
        vec = tile.top_l - tiles[0].top_l
        if vec.x < 1e-6 or vec.y < 1e-6:
            # All good the grid is not slanted
            break
    else:
        return "The grid is slanted", GridSetup()

    # Return the grid setup
    num_x, num_y = _find_grid_size(tiles, offset_x, offset_y)
    return None, GridSetup(
        length_x=length_x,
        length_y=length_y,
        offset_x=offset_x,
        offset_y=offset_y,
        num_x=num_x,
        num_y=num_y,
    )



"""Utility functions for building OME metadata from fractal-tasks-core models."""

from pathlib import Path

import zarr
from fractal_tasks_core.ngff.specs import (
    AcquisitionInPlate,
    ColumnInPlate,
    NgffPlateMeta,
    NgffWellMeta,
    Plate,
    RowInPlate,
    Well,
    WellInPlate,
)
from fractal_tasks_core.ngff.specs import ImageInWell as ImageInWellMeta

from fractal_converters_tools.tiled_image import TiledImage

alphabet = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"


def build_plate_meta(acquisitions: list[TiledImage], plate_name) -> NgffPlateMeta:
    """Build a plate metadata object from a list of acquisitions."""
    if len(acquisitions) == 0:
        raise ValueError("Empty list of acquisitions")

    _acquisition_ids = list({acq.acquisition_id for acq in acquisitions})
    acquisition_ids = []
    for acquisition_id in _acquisition_ids:
        acq_model = AcquisitionInPlate(
            id=acquisition_id,
            name=f"{plate_name}_id{acquisition_id}",
            maximumfieldcount=None,
            description=None,
        )
        acquisition_ids.append(acq_model)

    rows = []
    existing_rows = {acq.row for acq in acquisitions}
    for row_name in alphabet:
        if row_name in existing_rows:
            rows.append(RowInPlate(name=row_name))

    columns = []
    existing_columns = {acq.column for acq in acquisitions}
    for column_name in range(1, 100):
        if column_name in existing_columns:
            columns.append(ColumnInPlate(name=str(column_name)))

    wells = {}
    for row_id, row in enumerate(rows):
        for column_id, column in enumerate(columns):
            path = f"{row.name}/{column.name}"

            for acq in acquisitions:
                if acq.well_path == path and path not in wells:
                    wells[path] = WellInPlate(
                        path=path,
                        rowIndex=row_id,
                        columnIndex=column_id,
                    )
    wells_list = list(wells.values())

    plate = Plate(
        acquisitions=acquisition_ids,
        rows=rows,
        columns=columns,
        wells=wells_list,
        name=plate_name,
        version="0.4.0",
    )
    return NgffPlateMeta(plate=plate)


def build_well_meta(acquisitions: list[TiledImage]) -> dict[str, NgffWellMeta]:
    """Build a well metadata object from a list of acquisitions."""
    well_meta = {}

    for acq in acquisitions:
        if acq.well_path not in well_meta:
            well_meta[acq.well_path] = set()

        well_meta[acq.well_path].add(acq.acquisition_id)

    _well_meta = {}
    for path, wells in well_meta.items():
        images = []
        for acquisition_id in wells:
            images.append(
                ImageInWellMeta(acquisition=acquisition_id, path=str(acquisition_id))
            )

        _well_meta[path] = NgffWellMeta(well=Well(images=images, version="0.4.0"))
    return _well_meta


def initiate_ome_zarr_plate(
    store: str | Path,
    plate_name: str,
    acquisitions: list[TiledImage],
    overwrite: bool = False,
) -> None:
    """Create an OME-Zarr plate from a list of acquisitions."""
    plate_meta = build_plate_meta(acquisitions, plate_name)
    plate_wells_meta = build_well_meta(acquisitions)

    store = Path(store)
    if store.exists() and not overwrite:
        raise ValueError(
            f"Zarr file already exists at {store}. Set overwrite=True to overwrite."
        )

    plate_group = zarr.open_group(store, mode="w")
    plate_group.attrs.update(plate_meta.model_dump(exclude_none=True))

    for well_path, well_meta in plate_wells_meta.items():
        well_group = plate_group.create_group(well_path)
        well_group.attrs.update(well_meta.model_dump(exclude_none=True))



"""A module to represent an acquisition."""

from fractal_converters_tools.tile import Tile


class TiledImage:
    """A class to represent an acquisition."""

    def __init__(
        self,
        name: str,
        row: str,
        column: int,
        acquisition_id: int,
        tiles: list[Tile],
        channel_names: list[str] | None = None,
        num_levels: int = 5,
    ):
        """Initialize the acquisition."""
        self._name = name
        self._row = row
        self._column = column
        self._acquisition_id = acquisition_id
        self.tiles = tiles
        self._channel_names = channel_names
        self.num_levels = num_levels

    @property
    def acquisition_id(self) -> int:
        """Return the acquisition ID."""
        return self._acquisition_id

    @property
    def name(self) -> str:
        """Return the acquisition name."""
        return self._name

    @property
    def row(self) -> str:
        """Return the row."""
        return self._row

    @property
    def column(self) -> int:
        """Return the column."""
        return self._column

    @property
    def well_path(self) -> str:
        """Return the well path."""
        return f"{self._row}/{self._column}"

    @property
    def acquisition_path(self) -> str:
        """Return the acquisition path."""
        return f"{self.well_path}/{self._acquisition_id}"

    @property
    def channel_names(self) -> list[str]:
        """Return the channel names."""
        if self._channel_names is None:
            tile_channel_names = self.tiles[0].channel_names
            if tile_channel_names is None:
                raise ValueError(
                    "Channel names not found. Please provide channel names manually."
                )
            self._channel_names = tile_channel_names
        return self._channel_names

    @property
    def xy_scale(self) -> float:
        """Return the XY scale."""
        return self.tiles[0].xy_scale

    @property
    def z_scale(self) -> float:
        """Return the Z scale."""
        return self.tiles[0].z_scale



"""This module contains the classes to handle an abstact 5D (t, c, z, y, x) tile."""

from collections import namedtuple
from collections.abc import Callable
from dataclasses import dataclass
from enum import Enum
from typing import Protocol

import numpy as np
from dask.array.core import Array


def _find_prec(a: float | int) -> int:
    """Find the precision of a float."""
    if isinstance(a, int):
        return 0
    return len(str(a).split(".")[1])


def _round_ops(a: float | int, b: float | int, op: Callable) -> float | int:
    """Round and apply an operation to two numbers."""
    prec_a, prec_b = _find_prec(a), _find_prec(b)
    prec = max(prec_a, prec_b)
    return round(op(a, b), prec)


def _round_add(a: float | int, b: float | int) -> float | int:
    """Round and add two numbers."""
    return _round_ops(a, b, lambda x, y: x + y)


def _round_sub(a: float | int, b: float | int) -> float | int:
    """Round and subtract two numbers."""
    return _round_ops(a, b, lambda x, y: x - y)


@dataclass
class Vector:
    """Basic 5D vector class."""

    x: int | float
    y: int | float
    z: int | float = 0.0
    c: int = 0
    t: int = 0

    def __add__(self, other: "Vector") -> "Vector":
        """Add two vectors."""
        return Vector(
            _round_add(self.x, other.x),
            _round_add(self.y, other.y),
            _round_add(self.z, other.z),
            self.c + other.c,
            self.t + other.t,
        )

    def __sub__(self, other: "Vector") -> "Vector":
        """Subtract two vectors."""
        return Vector(
            _round_sub(self.x, other.x),
            _round_sub(self.y, other.y),
            _round_sub(self.z, other.z),
            self.c - other.c,
            self.t - other.t,
        )

    def __mul__(self, scalar: float) -> "Vector":
        """Multiply a vector by a scalar."""
        return Vector(
            self.x * scalar,
            self.y * scalar,
            self.z * scalar,
            int(self.c * scalar),
            int(self.t * scalar),
        )

    def normalize(self) -> "Vector":
        """Normalize the vector."""
        length = self.length()
        return Vector(
            self.x / length,
            self.y / length,
            self.z / length,
            int(self.c / length),
            int(self.t / length),
        )

    def length(self) -> float:
        """Compute the length of the vector."""
        return (self.x**2 + self.y**2 + self.z**2 + self.c**2 + self.t**2) ** 0.5

    def to_pixel_space(self, xy_scale: float = 1.0, z_scale: float = 1.0) -> "Vector":
        """Convert the vector to pixel space."""
        x = int(self.x * xy_scale)
        y = int(self.y * xy_scale)
        z = int(self.z * z_scale)
        return Vector(x, y, z=z, c=self.c, t=self.t)

    def to_real_space(self, xy_scale: float = 1.0, z_scale: float = 1.0) -> "Vector":
        """Convert the vector to real space."""
        x = self.x / xy_scale
        y = self.y / xy_scale
        z = self.z / z_scale
        return Vector(x, y, z=z, c=self.c, t=self.t)


@dataclass
class Point:
    """Basic 5D point class."""

    x: int | float
    y: int | float
    z: int | float = 0.0
    c: int = 0
    t: int = 0

    def __add__(self, other: Vector) -> "Point":
        """Add a vector to a point."""
        return Point(
            _round_add(self.x, other.x),
            _round_add(self.y, other.y),
            _round_add(self.z, other.z),
            self.c + other.c,
            self.t + other.t,
        )

    def __sub__(self, other: "Point") -> "Vector":
        """Subtract two points."""
        return Vector(
            _round_sub(self.x, other.x),
            _round_sub(self.y, other.y),
            _round_sub(self.z, other.z),
            self.c - other.c,
            self.t - other.t,
        )

    def to_pixel_space(self, xy_scale: float = 1.0, z_scale: float = 1.0) -> "Point":
        """Convert the point to pixel space."""
        x = int(self.x * xy_scale)
        y = int(self.y * xy_scale)
        z = int(self.z * z_scale)
        return Point(x, y, z=z, c=self.c, t=self.t)

    def to_real_space(self, xy_scale: float = 1.0, z_scale: float = 1.0) -> "Point":
        """Convert the point to real space."""
        x = self.x / xy_scale
        y = self.y / xy_scale
        z = self.z / z_scale
        return Point(x, y, z=z, c=self.c, t=self.t)


class TileLoader(Protocol):
    """Tile loader interface."""

    def __call__(self) -> np.ndarray | Array:
        """Load the tile data into a numpy array."""
        ...


class TileSpace(Enum):
    """Tile space enumeration."""

    PIXEL = "pixel"
    REAL = "real"


OriginDict = namedtuple(
    "OriginDict",
    [
        "x_micrometer_original",
        "y_micrometer_original",
        "z_micrometer_original",
        "t_original",
    ],
    defaults=[0.0, 0.0, 0.0, 0.0],
)


class Tile:
    """5D tile class.

    The tile is defined by two points: the top-left and bottom-right corners.

    The origin attribute is used to keep track of the original tile position when
    moving
    """

    def __init__(
        self,
        top_l: Point,
        diag: Vector,
        origin: OriginDict | None = None,
        space: TileSpace = TileSpace.REAL,
        data_loader: TileLoader | None = None,
        channel_names: list[str] | None = None,
        xy_scale: float = 1.0,
        z_scale: float = 1.0,
    ):
        """Initialize the tile with the top-left corner and the diagonal vector."""
        self._top_l = top_l
        self._diag = diag

        if origin is None:
            self._origin = OriginDict(
                x_micrometer_original=top_l.x, y_micrometer_original=top_l.y
            )
        else:
            self._origin = origin

        self._data_loader = data_loader
        self._space = space
        self._channel_names = channel_names
        self._xy_scale = xy_scale
        self._z_scale = z_scale

    def __repr__(self) -> str:
        """String representation of the tile."""
        x, y, z, c, t = (
            self.top_l.x,
            self.top_l.y,
            self.top_l.z,
            self.top_l.c,
            self.top_l.t,
        )
        s_x, s_y, s_z, s_c, s_t = (
            self.diag.x,
            self.diag.y,
            self.diag.z,
            self.diag.c,
            self.diag.t,
        )
        o_x, o_y = (
            self.origin.x_micrometer_original,
            self.origin.y_micrometer_original,
        )
        return (
            f"Tile(xyzct=({x}, {y}, {z}, {c}, {t}), "
            f"diag=({s_x}, {s_y}, {s_z}, {s_c}, {s_t})), "
            f"origin=({o_x}, {o_y}), "
            f"space={self.space})"
        )

    @property
    def top_l(self) -> Point:
        """Return the top-left corner of the tile."""
        return self._top_l

    @property
    def diag(self) -> Vector:
        """Return the diagonal vector of the tile."""
        return self._diag

    @property
    def bot_r(self) -> Point:
        """Return the bottom-right corner of the tile."""
        return self._top_l + self._diag

    @property
    def origin(self) -> OriginDict:
        """Return the origin reference of the tile."""
        return self._origin

    @property
    def space(self) -> TileSpace:
        """Return the space of the tile."""
        return self._space

    @property
    def channel_names(self) -> list[str] | None:
        """Return the channel names of the tile."""
        return self._channel_names

    @property
    def xy_scale(self) -> float:
        """Return the XY scale of the tile."""
        return self._xy_scale

    @property
    def z_scale(self) -> float:
        """Return the Z scale of the tile."""
        return self._z_scale

    @classmethod
    def from_points(
        cls,
        top_l: Point,
        bot_r: Point,
        origin: OriginDict | None = None,
        space: TileSpace = TileSpace.REAL,
        data_loader: TileLoader | None = None,
        channel_names: list[str] | None = None,
        xy_scale: float = 1.0,
        z_scale: float = 1.0,
    ):
        """Create a tile from two points (top-left and bottom-right corners)."""
        diag = bot_r - top_l
        return cls(
            top_l=top_l,
            diag=diag,
            origin=origin,
            space=space,
            data_loader=data_loader,
            channel_names=channel_names,
            xy_scale=xy_scale,
            z_scale=z_scale,
        )

    def derive_from_diag(self, top_l: Point, diag: Vector) -> "Tile":
        """Create a new tile keeping the origin."""
        return Tile(
            top_l,
            diag,
            origin=self._origin,
            data_loader=self._data_loader,
            space=self._space,
            channel_names=self._channel_names,
            xy_scale=self._xy_scale,
            z_scale=self._z_scale,
        )

    def derive_from_points(self, top_l: Point, bot_r: Point) -> "Tile":
        """Create a new tile keeping the origin."""
        diag = bot_r - top_l
        return Tile(
            top_l,
            diag,
            origin=self._origin,
            data_loader=self._data_loader,
            space=self._space,
            channel_names=self._channel_names,
            xy_scale=self._xy_scale,
            z_scale=self._z_scale,
        )

    def move_by(self, vec: Vector) -> "Tile":
        """Move the tile by a vector keeping the origin reference."""
        return self.derive_from_diag(self.top_l + vec, self.diag)

    def move_to(self, point: Point) -> "Tile":
        """Move the tile to a new point."""
        return self.derive_from_points(point, point + self.diag)

    def reset_origin(self) -> "Tile":
        """Reset the origin reference of the tile to the current position."""
        new_origin = OriginDict(
            x_micrometer_original=self.top_l.x, y_micrometer_original=self.top_l.y
        )
        return Tile(
            self.top_l,
            self.diag,
            origin=new_origin,
            data_loader=self._data_loader,
            space=self._space,
            channel_names=self._channel_names,
            xy_scale=self._xy_scale,
            z_scale=self._z_scale,
        )

    def to_pixel_space(self) -> "Tile":
        """Convert the tile to pixel space."""
        if self.space == TileSpace.PIXEL:
            raise ValueError("Tile is already in pixel space")
        top_l = self.top_l.to_pixel_space(
            xy_scale=1 / self.xy_scale, z_scale=1 / self.z_scale
        )
        diag = self.diag.to_pixel_space(
            xy_scale=1 / self.xy_scale, z_scale=1 / self.z_scale
        )
        return Tile(
            top_l,
            diag,
            origin=self._origin,
            data_loader=self._data_loader,
            space=TileSpace.PIXEL,
            channel_names=self._channel_names,
            xy_scale=self._xy_scale,
            z_scale=self._z_scale,
        )

    def to_real_space(self) -> "Tile":
        """Convert the tile to real space."""
        if self.space == TileSpace.REAL:
            raise ValueError("Tile is already in real space")
        top_l = self.top_l.to_real_space(
            xy_scale=1 / self.xy_scale, z_scale=1 / self.z_scale
        )
        diag = self.diag.to_real_space(
            xy_scale=1 / self.xy_scale, z_scale=1 / self.z_scale
        )
        return Tile(
            top_l,
            diag,
            origin=self._origin,
            data_loader=self._data_loader,
            space=TileSpace.REAL,
            channel_names=self._channel_names,
            xy_scale=self._xy_scale,
            z_scale=self._z_scale,
        )

    def is_coplanar(self, bbox: "Tile", z_tol: float = 1e-6) -> bool:
        """Check if two tiles are coplanar on the XY plane."""
        if abs(self.top_l.z - bbox.top_l.z) > z_tol:
            return False

        if self.top_l.c != bbox.top_l.c:
            return False

        if self.top_l.t != bbox.top_l.t:
            return False

        return True

    def cornersXY(self) -> list[Point]:
        """Return the 4 corners of the tiles box in the top-XY plane."""
        corners = [
            (self.top_l.x, self.top_l.y),
            (self.top_l.x, self.bot_r.y),
            (self.bot_r.x, self.bot_r.y),
            (self.bot_r.x, self.top_l.y),
        ]

        return [
            Point(x, y, self.top_l.z, self.top_l.c, self.top_l.t) for x, y in corners
        ]

    def areaXY(self) -> float:
        """BBBox area in the XY plane."""
        return self.diag.x * self.diag.y

    def intersection_area_XY(self, other: "Tile") -> float:
        """Compute the intersection of two tiles boxes on the XY plane."""
        # Check if the bounding boxes are coplanar
        if not self.is_coplanar(other):
            raise ValueError("Bounding boxes are not coplanar")

        # Compute the intersection
        min_x = max(self.top_l.x, other.top_l.x)
        min_y = max(self.top_l.y, other.top_l.y)

        max_x = min(self.bot_r.x, other.bot_r.x)
        max_y = min(self.bot_r.y, other.bot_r.y)

        if min_x > max_x or min_y > max_y:
            return 0

        return (max_x - min_x) * (max_y - min_y)

    def iouXY(self, other: "Tile") -> float:
        """Compute the intersection over union of tiles in the XY plane."""
        if not self._is_overlappingXY(other):
            return 0

        area_inter = self.intersection_area_XY(other)
        if area_inter <= 0:
            return 0

        vol1 = self.areaXY()
        vol2 = other.areaXY()
        union = vol1 + vol2 - area_inter
        if union <= 0:
            raise ValueError("Union is less than 0")

        return area_inter / union

    def _is_overlappingXY(self, bbox: "Tile") -> bool:
        if self.top_l.x > bbox.bot_r.x or self.bot_r.x < bbox.top_l.x:
            return False
        if self.top_l.y > bbox.bot_r.y or self.bot_r.y < bbox.top_l.y:
            return False
        return True

    def is_overlappingXY(self, bbox: "Tile", eps: float = 1e-6) -> bool:
        """Check if two bounding boxes are overlapping."""
        return self._is_overlappingXY(bbox) and self.iouXY(bbox) > eps

    def load(self) -> np.ndarray | Array:
        """Load the tile data."""
        if self._data_loader is None:
            raise ValueError("No data loader provided.")
        return self._data_loader()

    def shape(self) -> dict[str, int]:
        """Return the shape of the tile."""
        if self.space == TileSpace.REAL:
            raise ValueError("Tile is in real space, cannot return shape")

        return {
            "t": int(self.diag.t),
            "c": int(self.diag.c),
            "z": int(self.diag.z),
            "y": int(self.diag.y),
            "x": int(self.diag.x),
        }

    def slices(self) -> dict[str, slice]:
        """Return the slices of the tile."""
        if self.space == TileSpace.REAL:
            raise ValueError("Tile is in real space, cannot return slices")

        return {
            "t": slice(self.top_l.t, self.bot_r.t),
            "c": slice(self.top_l.c, self.bot_r.c),
            "z": slice(self.top_l.z, self.bot_r.z),
            "y": slice(self.top_l.y, self.bot_r.y),
            "x": slice(self.top_l.x, self.bot_r.x),
        }



"""Utilities for working with microplates."""

STANDARD_PLATES_LAYOUTS = {
    "24-well": {
        "rows": 4,
        "columns": 6,
    },
    "48-well": {
        "rows": 6,
        "columns": 8,
    },
    "96-well": {
        "rows": 8,
        "columns": 12,
    },
    "384-well": {
        "rows": 16,
        "columns": 24,
    },
}

STANDARD_ROWS_NAMES = "ABCDEFGHIJKLMNOP"


def get_row_column(well_id: int, layout: str) -> tuple[str, int]:
    """Get row and column from well id."""
    if layout not in STANDARD_PLATES_LAYOUTS:
        raise ValueError(f"Layout {layout} not found.")

    layout_dict = STANDARD_PLATES_LAYOUTS[layout]
    num_columns = layout_dict["columns"]
    well_id -= 1

    row = well_id // num_columns
    column = well_id % num_columns + 1

    if row >= layout_dict["rows"]:
        raise ValueError(f"Well id {well_id} is out of bounds for layout {layout}.")

    row_str = STANDARD_ROWS_NAMES[row]
    return row_str, column



"""Tooling to build ome-zarr HCS plate converters."""

from importlib.metadata import PackageNotFoundError, version

from fractal_converters_tools.tiled_image import TiledImage
from fractal_converters_tools.ome_plate_meta import initiate_ome_zarr_plate

try:
    __version__ = version("fractal-converters-tools")
except PackageNotFoundError:
    __version__ = "uninstalled"
__author__ = "Lorenzo Cerrone"
__email__ = "lorenzo.cerrone@uzh.ch"

__all__ = ["TiledImage", "initiate_ome_zarr_plate"]



"""Utility functions for the tiles module."""

import copy
from typing import Literal

import numpy as np

from fractal_converters_tools.grid_utils import GridSetup, check_if_regular_grid
from fractal_converters_tools.tile import Point, Tile, Vector


def check_tiles_coplanar(tiles: list[Tile]) -> bool:
    """Check if all the Tiles are coplanar on the XY plane."""
    if len(tiles) == 0:
        return True

    return all(tiles[0].is_coplanar(tile) for tile in tiles)


def sort_tiles_by_distance(
    tiles: list[Tile], check_coplanar: bool = False
) -> list[Tile]:
    """Sort a list of tiles by distance from the origin."""
    # Check if the tiles are coplanar on the XY plane
    if check_coplanar and not check_tiles_coplanar(tiles):
        raise ValueError("Tiles are not coplanar")

    min_x = min([tile.top_l.x for tile in tiles])
    min_y = min([tile.top_l.y for tile in tiles])
    min_point = Point(
        min_x, min_y, tiles[0].top_l.z, tiles[0].top_l.c, tiles[0].top_l.t
    )
    return sorted(tiles, key=lambda x: (x.top_l - min_point).length())


def remove_tiles_offset(tiles: list[Tile]) -> list[Tile]:
    """Remove the offset from a list of tiles."""
    sorted_tiles = sort_tiles_by_distance(tiles)
    origin = tiles[0].top_l
    offset_vector = Vector(-origin.x, -origin.y, z=-origin.z, c=-origin.c, t=-origin.t)
    return [tile.move_by(vec=offset_vector) for tile in sorted_tiles]


def tiles_to_pixel_space(tiles: list[Tile]) -> list[Tile]:
    """Convert a list of tiles from real space to pixel space."""
    return [tile.to_pixel_space() for tile in tiles]


def tiles_to_real_space(tiles: list[Tile]) -> list[Tile]:
    """Convert a list of tiles from pixel space to real space."""
    return [tile.to_real_space() for tile in tiles]


def swap_xy_tiles(tiles: list[Tile]):
    """Swap x and y of the tiles."""
    swapped_tiles = []
    for tile in tiles:
        top_l = Point(
            tile.top_l.y, tile.top_l.x, z=tile.top_l.z, c=tile.top_l.c, t=tile.top_l.t
        )
        diag = Vector(
            tile.diag.y, tile.diag.x, z=tile.diag.z, c=tile.diag.c, t=tile.diag.t
        )
        swapped_tiles.append(tile.derive_from_diag(top_l, diag))
    return swapped_tiles


def invert_x_tiles(tiles: list[Tile]) -> list[Tile]:
    """Invert the x coordinate of the tiles."""
    inverted_tiles = []
    for tile in tiles:
        top_l = Point(
            -tile.top_l.x, tile.top_l.y, z=tile.top_l.z, c=tile.top_l.c, t=tile.top_l.t
        )
        inverted_tiles.append(tile.derive_from_diag(top_l, tile.diag))
    return inverted_tiles


def invert_y_tiles(tiles: list[Tile]) -> list[Tile]:
    """Invert the y coordinate of the tiles."""
    inverted_tiles = []
    for tile in tiles:
        top_l = Point(
            tile.top_l.x, -tile.top_l.y, z=tile.top_l.z, c=tile.top_l.c, t=tile.top_l.t
        )
        inverted_tiles.append(tile.derive_from_diag(top_l, tile.diag))
    return inverted_tiles


def reset_tiles_origin(tiles: list[Tile]) -> list[Tile]:
    """Reset the tiles to their original position."""
    return [tile.reset_origin() for tile in tiles]


def _remove_tile_XY_overalap(
    ref_tile: Tile, query_tile: Tile, speed=1.0, eps: float = 1e-6
) -> Tile:
    """Move the query_tile to remove the overlap with the ref_tile."""
    if speed <= 0 or speed > 1:
        raise ValueError("Speed must be in the range (0, 1]")

    lenghts = []
    vectors = []
    for corner in ref_tile.cornersXY():
        vec = corner - query_tile.top_l
        moved_bbox = query_tile.move_by(vec)
        iou = moved_bbox.iouXY(ref_tile)
        if iou < eps:
            lenghts.append(vec.length())
            vectors.append(vec)

    min_idx = np.argmin(lenghts)
    best_vector = vectors[min_idx]
    best_vector = best_vector * speed
    best_moved_bbox = query_tile.move_by(best_vector)
    return best_moved_bbox


def resolve_random_tiles_overlap(
    tiles: list[Tile], sort_list: bool = True, eps: float = 1e-6
) -> list[Tile]:
    """Remove the overlap from any list of tiles."""
    tiles = copy.deepcopy(tiles)

    if sort_list:
        tiles = sort_tiles_by_distance(tiles)

    n_overlap = np.inf
    while n_overlap > 0:
        n_overlap = 0
        for i in range(len(tiles)):
            tile = tiles[i]
            for j in range(i + 1, len(tiles)):
                query_tile = tiles[j]
                if tile.is_overlappingXY(query_tile, eps=eps):
                    bbox_no = _remove_tile_XY_overalap(tile, query_tile, speed=1)
                    tiles[j] = bbox_no
                    n_overlap += 1
                    break
    return tiles


def resolve_grid_tiles_overlap(
    tiles: list[Tile], grid_setup: GridSetup, eps: float = 1e-6
) -> list[Tile]:
    """Remove overlap from a list of tiles that follow a regular grid."""
    tiles = sort_tiles_by_distance(tiles)

    z, c, t = tiles[0].top_l.z, tiles[0].top_l.c, tiles[0].top_l.t

    output_tiles = []
    for i in range(grid_setup.num_x):
        for j in range(grid_setup.num_y):
            # X-Y position in the input grid
            x_in = i * grid_setup.offset_x
            y_in = j * grid_setup.offset_y

            # X-Y position in the output grid
            x_out = i * grid_setup.length_x
            y_out = j * grid_setup.length_y

            # Find if a bounding box is close to the (x_in, y_in) position
            point = Point(x_in, y_in, z=z, c=c, t=t)
            distances = [(point - bbox.top_l).length() for bbox in tiles]
            min_dist = np.min(distances)
            closest_bbox = tiles[np.argmin(distances)]

            if min_dist < eps:
                # Move the bounding box to the (x_out, y_out) position
                top_l = Point(x_out, y_out, z=z, c=c, t=t)
                new_tile = closest_bbox.derive_from_diag(top_l, diag=closest_bbox.diag)
                output_tiles.append(new_tile)
    return output_tiles


def _resolve_auto_mode(tiles: list[Tile]) -> list[Tile]:
    """Resolve the overlap of a list of tiles."""
    error_message_or_none, grid_setup = check_if_regular_grid(tiles)
    if error_message_or_none is None:
        return resolve_grid_tiles_overlap(tiles, grid_setup)
    return resolve_random_tiles_overlap(tiles)

def _resolve_grid_mode(tiles: list[Tile]) -> list[Tile]:
    """Resolve the overlap of a list of tiles."""
    error_message_or_none, grid_setup = check_if_regular_grid(tiles)
    if error_message_or_none is not None:
        raise ValueError(
            "The input tiles are not on a regular grid "
            f"because: {error_message_or_none}. Please set 'mode=free'."
        )
    return resolve_grid_tiles_overlap(tiles, grid_setup)

def _resolve_free_mode(tiles: list[Tile]) -> list[Tile]:
    """Resolve the overlap of a list of tiles."""
    return resolve_random_tiles_overlap(tiles)

def resolve_tiles_overlap(
    tiles: list[Tile],
    mode: Literal["auto", "grid", "free", "none"] = "auto",
) -> list[Tile]:
    """Remove the overlap from any list of tiles."""
    if mode not in ["auto", "grid", "free", "none"]:
        raise ValueError("Mode must be 'auto', 'grid', 'free', or 'none'")

    match mode:
        case "auto":
            return _resolve_auto_mode(tiles)
        case "grid":
            return _resolve_grid_mode(tiles)
        case "free":
            return _resolve_free_mode(tiles)
        case "none":
            return tiles

def standard_stitching_pipe(tiles: list[Tile], 
                            mode: Literal["auto", "grid", "free", "none"] = "auto",
                            swap_xy: bool = False,
                            invert_x: bool = False,
                            invert_y: bool = False) -> list[Tile]:
    """Standard stitching pipe for a list of tiles."""
    tiles = copy.deepcopy(tiles)
    if swap_xy:
        tiles = swap_xy_tiles(tiles)
    if invert_x:
        tiles = invert_x_tiles(tiles)
    if invert_y:
        tiles = invert_y_tiles(tiles)
    
    if any([swap_xy, invert_x, invert_y]):
        tiles = remove_tiles_offset(tiles)
    
    tiles = sort_tiles_by_distance(tiles)
    tiles = remove_tiles_offset(tiles)
    tiles = resolve_tiles_overlap(tiles, mode=mode)
    tiles = tiles_to_pixel_space(tiles)
    return tiles
